package com.sun.org.apache.xerces.internal.impl.xs.traversers;

import com.sun.org.apache.xerces.internal.impl.xs.XSElementDecl;
import com.sun.org.apache.xerces.internal.impl.xs.XSParticleDecl;
import com.sun.org.apache.xerces.internal.impl.xs.XSWildcardDecl;
import com.sun.org.apache.xerces.internal.xs.XSAttributeUse;
import com.sun.org.apache.xerces.internal.impl.xs.XSAttributeUseImpl;
import com.sun.org.apache.xerces.internal.impl.xs.XSAttributeGroupDecl;
import com.sun.org.apache.xerces.internal.xs.XSObjectList;
import com.sun.org.apache.xerces.internal.impl.dv.util.Base64;
import com.sun.org.apache.xerces.internal.impl.dv.xs.XSSimpleTypeDecl;
import com.sun.org.apache.xerces.internal.impl.xs.XSComplexTypeDecl;
import com.sun.org.apache.xerces.internal.xs.XSSimpleTypeDefinition;
import com.sun.org.apache.xerces.internal.xni.NamespaceContext;
import com.sun.org.apache.xerces.internal.impl.dv.InvalidDatatypeValueException;
import com.sun.org.apache.xerces.internal.xni.QName;
import com.sun.org.apache.xerces.internal.util.NamespaceSupport;
import com.sun.org.apache.xerces.internal.impl.xs.util.XSObjectListImpl;
import com.sun.org.apache.xerces.internal.impl.xs.util.XInt;
import com.sun.org.apache.xerces.internal.xs.XSTypeDefinition;
import java.util.List;
import com.sun.org.apache.xerces.internal.impl.xs.SchemaGrammar;
import java.util.ArrayList;
import com.sun.org.apache.xerces.internal.impl.xs.SchemaSymbols;
import com.sun.org.apache.xerces.internal.util.DOMUtil;
import com.sun.org.apache.xerces.internal.impl.xs.XSAnnotationImpl;
import org.w3c.dom.Element;
import java.util.Locale;
import com.sun.org.apache.xerces.internal.impl.dv.XSFacets;
import com.sun.org.apache.xerces.internal.impl.dv.XSSimpleType;
import com.sun.org.apache.xerces.internal.impl.validation.ValidationState;
import com.sun.org.apache.xerces.internal.util.SymbolTable;

abstract class XSDAbstractTraverser
{
    protected static final String NO_NAME = "(no name)";
    protected static final int NOT_ALL_CONTEXT = 0;
    protected static final int PROCESSING_ALL_EL = 1;
    protected static final int GROUP_REF_WITH_ALL = 2;
    protected static final int CHILD_OF_GROUP = 4;
    protected static final int PROCESSING_ALL_GP = 8;
    protected XSDHandler fSchemaHandler;
    protected SymbolTable fSymbolTable;
    protected XSAttributeChecker fAttrChecker;
    protected boolean fValidateAnnotations;
    ValidationState fValidationState;
    private static final XSSimpleType fQNameDV;
    private StringBuilder fPattern;
    private final XSFacets xsFacets;
    
    XSDAbstractTraverser(XSDHandler handler, XSAttributeChecker attrChecker) {
        this.fSchemaHandler = null;
        this.fSymbolTable = null;
        this.fAttrChecker = null;
        this.fValidateAnnotations = false;
        this.fValidationState = new ValidationState();
        this.fPattern = new StringBuilder();
        this.xsFacets = new XSFacets();
        this.fSchemaHandler = handler;
        this.fAttrChecker = attrChecker;
    }
    
    void reset(SymbolTable symbolTable, boolean validateAnnotations, Locale locale) {
        this.fSymbolTable = symbolTable;
        this.fValidateAnnotations = validateAnnotations;
        this.fValidationState.setExtraChecking(false);
        this.fValidationState.setSymbolTable(symbolTable);
        this.fValidationState.setLocale(locale);
    }
    
    XSAnnotationImpl traverseAnnotationDecl(Element annotationDecl, Object[] parentAttrs, boolean isGlobal, XSDocumentInfo schemaDoc) {
        Object[] attrValues = this.fAttrChecker.checkAttributes(annotationDecl, isGlobal, schemaDoc);
        this.fAttrChecker.returnAttrArray(attrValues, schemaDoc);
        String contents = DOMUtil.getAnnotation(annotationDecl);
        Element child = DOMUtil.getFirstChildElement(annotationDecl);
        if (child != null) {
            do {
                String name = DOMUtil.getLocalName(child);
                if (!name.equals(SchemaSymbols.ELT_APPINFO) && !name.equals(SchemaSymbols.ELT_DOCUMENTATION)) {
                    this.reportSchemaError("src-annotation", new Object[] { name }, child);
                }
                else {
                    attrValues = this.fAttrChecker.checkAttributes(child, true, schemaDoc);
                    this.fAttrChecker.returnAttrArray(attrValues, schemaDoc);
                }
                child = DOMUtil.getNextSiblingElement(child);
            } while (child != null);
        }
        if (contents == null) {
            return null;
        }
        SchemaGrammar grammar = this.fSchemaHandler.getGrammar(schemaDoc.fTargetNamespace);
        List<String> annotationLocalAttrs = (ArrayList)parentAttrs[XSAttributeChecker.ATTIDX_NONSCHEMA];
        if (annotationLocalAttrs != null && !annotationLocalAttrs.isEmpty()) {
            StringBuilder localStrBuffer = new StringBuilder(64);
            localStrBuffer.append(" ");
            int i = 0;
            while (i < annotationLocalAttrs.size()) {
                String rawname = (String)annotationLocalAttrs.get(i++);
                int colonIndex = rawname.indexOf(58);
                String prefix;
                String localpart;
                if (colonIndex == -1) {
                    prefix = "";
                    localpart = rawname;
                }
                else {
                    prefix = rawname.substring(0, colonIndex);
                    localpart = rawname.substring(colonIndex + 1);
                }
                String uri = schemaDoc.fNamespaceSupport.getURI(this.fSymbolTable.addSymbol(prefix));
                if (annotationDecl.getAttributeNS(uri, localpart).length() != 0) {
                    ++i;
                }
                else {
                    localStrBuffer.append(rawname).append("=\"");
                    String value = (String)annotationLocalAttrs.get(i++);
                    value = processAttValue(value);
                    localStrBuffer.append(value).append("\" ");
                }
            }
            StringBuilder contentBuffer = new StringBuilder(contents.length() + localStrBuffer.length());
            int annotationTokenEnd = contents.indexOf(SchemaSymbols.ELT_ANNOTATION);
            if (annotationTokenEnd == -1) {
                return null;
            }
            annotationTokenEnd += SchemaSymbols.ELT_ANNOTATION.length();
            contentBuffer.append(contents.substring(0, annotationTokenEnd));
            contentBuffer.append(localStrBuffer.toString());
            contentBuffer.append(contents.substring(annotationTokenEnd, contents.length()));
            String annotation = contentBuffer.toString();
            if (this.fValidateAnnotations) {
                schemaDoc.addAnnotation(new XSAnnotationInfo(annotation, annotationDecl));
            }
            return new XSAnnotationImpl(annotation, grammar);
        }
        else {
            if (this.fValidateAnnotations) {
                schemaDoc.addAnnotation(new XSAnnotationInfo(contents, annotationDecl));
            }
            return new XSAnnotationImpl(contents, grammar);
        }
    }
    
    XSAnnotationImpl traverseSyntheticAnnotation(Element annotationParent, String initialContent, Object[] parentAttrs, boolean isGlobal, XSDocumentInfo schemaDoc) {
        String contents = initialContent;
        SchemaGrammar grammar = this.fSchemaHandler.getGrammar(schemaDoc.fTargetNamespace);
        List<String> annotationLocalAttrs = (ArrayList)parentAttrs[XSAttributeChecker.ATTIDX_NONSCHEMA];
        if (annotationLocalAttrs != null && !annotationLocalAttrs.isEmpty()) {
            StringBuilder localStrBuffer = new StringBuilder(64);
            localStrBuffer.append(" ");
            int i = 0;
            while (i < annotationLocalAttrs.size()) {
                String rawname = (String)annotationLocalAttrs.get(i++);
                int colonIndex = rawname.indexOf(58);
                String prefix;
                if (colonIndex == -1) {
                    prefix = "";
                    String localpart = rawname;
                }
                else {
                    prefix = rawname.substring(0, colonIndex);
                    String localpart = rawname.substring(colonIndex + 1);
                }
                String uri = schemaDoc.fNamespaceSupport.getURI(this.fSymbolTable.addSymbol(prefix));
                localStrBuffer.append(rawname).append("=\"");
                String value = (String)annotationLocalAttrs.get(i++);
                value = processAttValue(value);
                localStrBuffer.append(value).append("\" ");
            }
            StringBuilder contentBuffer = new StringBuilder(contents.length() + localStrBuffer.length());
            int annotationTokenEnd = contents.indexOf(SchemaSymbols.ELT_ANNOTATION);
            if (annotationTokenEnd == -1) {
                return null;
            }
            annotationTokenEnd += SchemaSymbols.ELT_ANNOTATION.length();
            contentBuffer.append(contents.substring(0, annotationTokenEnd));
            contentBuffer.append(localStrBuffer.toString());
            contentBuffer.append(contents.substring(annotationTokenEnd, contents.length()));
            String annotation = contentBuffer.toString();
            if (this.fValidateAnnotations) {
                schemaDoc.addAnnotation(new XSAnnotationInfo(annotation, annotationParent));
            }
            return new XSAnnotationImpl(annotation, grammar);
        }
        else {
            if (this.fValidateAnnotations) {
                schemaDoc.addAnnotation(new XSAnnotationInfo(contents, annotationParent));
            }
            return new XSAnnotationImpl(contents, grammar);
        }
    }
    
    XSDAbstractTraverser.FacetInfo traverseFacets(Element content, XSTypeDefinition typeDef, XSSimpleType baseValidator, XSDocumentInfo schemaDoc) {
        short facetsPresent = 0;
        short facetsFixed = 0;
        boolean hasQName = this.containsQName(baseValidator);
        List<String> enumData = null;
        XSObjectListImpl enumAnnotations = null;
        XSObjectListImpl patternAnnotations = null;
        List<NamespaceContext> enumNSDecls = hasQName ? new ArrayList() : null;
        int currentFacet = 0;
        this.xsFacets.reset();
        boolean seenPattern = false;
        Element contextNode = (Element)content.getParentNode();
        boolean hasLengthFacet = false;
        boolean hasMinLengthFacet = false;
        boolean hasMaxLengthFacet = false;
        while (content != null) {
            Object[] attrs = null;
            String facet = DOMUtil.getLocalName(content);
            if (facet.equals(SchemaSymbols.ELT_ENUMERATION)) {
                attrs = this.fAttrChecker.checkAttributes(content, false, schemaDoc, hasQName);
                String enumVal = (String)attrs[XSAttributeChecker.ATTIDX_VALUE];
                if (enumVal == null) {
                    this.reportSchemaError("s4s-att-must-appear", new Object[] { SchemaSymbols.ELT_ENUMERATION, SchemaSymbols.ATT_VALUE }, content);
                    this.fAttrChecker.returnAttrArray(attrs, schemaDoc);
                    content = DOMUtil.getNextSiblingElement(content);
                    continue;
                }
                NamespaceSupport nsDecls = (NamespaceSupport)attrs[XSAttributeChecker.ATTIDX_ENUMNSDECLS];
                if (baseValidator.getVariety() == 1 && baseValidator.getPrimitiveKind() == 20) {
                    schemaDoc.fValidationContext.setNamespaceSupport(nsDecls);
                    Object notation = null;
                    try {
                        QName temp = (QName)XSDAbstractTraverser.fQNameDV.validate(enumVal, schemaDoc.fValidationContext, null);
                        notation = this.fSchemaHandler.getGlobalDecl(schemaDoc, 6, temp, content);
                    }
                    catch (InvalidDatatypeValueException ex) {
                        this.reportSchemaError(ex.getKey(), ex.getArgs(), content);
                    }
                    if (notation == null) {
                        this.fAttrChecker.returnAttrArray(attrs, schemaDoc);
                        content = DOMUtil.getNextSiblingElement(content);
                        continue;
                    }
                    schemaDoc.fValidationContext.setNamespaceSupport(schemaDoc.fNamespaceSupport);
                }
                if (enumData == null) {
                    enumData = new ArrayList();
                    enumAnnotations = new XSObjectListImpl();
                }
                enumData.add(enumVal);
                enumAnnotations.addXSObject(null);
                if (hasQName) {
                    enumNSDecls.add(nsDecls);
                }
                Element child3 = DOMUtil.getFirstChildElement(content);
                if (child3 != null && DOMUtil.getLocalName(child3).equals(SchemaSymbols.ELT_ANNOTATION)) {
                    enumAnnotations.addXSObject(enumAnnotations.getLength() - 1, this.traverseAnnotationDecl(child3, attrs, false, schemaDoc));
                    child3 = DOMUtil.getNextSiblingElement(child3);
                }
                else {
                    String text2 = DOMUtil.getSyntheticAnnotation(content);
                    if (text2 != null) {
                        enumAnnotations.addXSObject(enumAnnotations.getLength() - 1, this.traverseSyntheticAnnotation(content, text2, attrs, false, schemaDoc));
                    }
                }
                if (child3 != null) {
                    this.reportSchemaError("s4s-elt-must-match.1", new Object[] { "enumeration", "(annotation?)", DOMUtil.getLocalName(child3) }, child3);
                }
            }
            else if (facet.equals(SchemaSymbols.ELT_PATTERN)) {
                attrs = this.fAttrChecker.checkAttributes(content, false, schemaDoc);
                String patternVal = (String)attrs[XSAttributeChecker.ATTIDX_VALUE];
                if (patternVal == null) {
                    this.reportSchemaError("s4s-att-must-appear", new Object[] { SchemaSymbols.ELT_PATTERN, SchemaSymbols.ATT_VALUE }, content);
                    this.fAttrChecker.returnAttrArray(attrs, schemaDoc);
                    content = DOMUtil.getNextSiblingElement(content);
                    continue;
                }
                seenPattern = true;
                if (this.fPattern.length() == 0) {
                    this.fPattern.append(patternVal);
                }
                else {
                    this.fPattern.append("|");
                    this.fPattern.append(patternVal);
                }
                Element child2 = DOMUtil.getFirstChildElement(content);
                if (child2 != null && DOMUtil.getLocalName(child2).equals(SchemaSymbols.ELT_ANNOTATION)) {
                    if (patternAnnotations == null) {
                        patternAnnotations = new XSObjectListImpl();
                    }
                    patternAnnotations.addXSObject(this.traverseAnnotationDecl(child2, attrs, false, schemaDoc));
                    child2 = DOMUtil.getNextSiblingElement(child2);
                }
                else {
                    String text = DOMUtil.getSyntheticAnnotation(content);
                    if (text != null) {
                        if (patternAnnotations == null) {
                            patternAnnotations = new XSObjectListImpl();
                        }
                        patternAnnotations.addXSObject(this.traverseSyntheticAnnotation(content, text, attrs, false, schemaDoc));
                    }
                }
                if (child2 != null) {
                    this.reportSchemaError("s4s-elt-must-match.1", new Object[] { "pattern", "(annotation?)", DOMUtil.getLocalName(child2) }, child2);
                }
            }
            else {
                if (facet.equals(SchemaSymbols.ELT_MINLENGTH)) {
                    currentFacet = 2;
                }
                else if (facet.equals(SchemaSymbols.ELT_MAXLENGTH)) {
                    currentFacet = 4;
                }
                else if (facet.equals(SchemaSymbols.ELT_MAXEXCLUSIVE)) {
                    currentFacet = 64;
                }
                else if (facet.equals(SchemaSymbols.ELT_MAXINCLUSIVE)) {
                    currentFacet = 32;
                }
                else if (facet.equals(SchemaSymbols.ELT_MINEXCLUSIVE)) {
                    currentFacet = 128;
                }
                else if (facet.equals(SchemaSymbols.ELT_MININCLUSIVE)) {
                    currentFacet = 256;
                }
                else if (facet.equals(SchemaSymbols.ELT_TOTALDIGITS)) {
                    currentFacet = 512;
                }
                else if (facet.equals(SchemaSymbols.ELT_FRACTIONDIGITS)) {
                    currentFacet = 1024;
                }
                else if (facet.equals(SchemaSymbols.ELT_WHITESPACE)) {
                    currentFacet = 16;
                }
                else if (facet.equals(SchemaSymbols.ELT_LENGTH)) {
                    currentFacet = 1;
                }
                else {
                    break;
                }
                attrs = this.fAttrChecker.checkAttributes(content, false, schemaDoc);
                if ((facetsPresent & currentFacet) != 0x0) {
                    this.reportSchemaError("src-single-facet-value", new Object[] { facet }, content);
                    this.fAttrChecker.returnAttrArray(attrs, schemaDoc);
                    content = DOMUtil.getNextSiblingElement(content);
                    continue;
                }
                if (attrs[XSAttributeChecker.ATTIDX_VALUE] == null) {
                    if (content.getAttributeNodeNS(null, "value") == null) {
                        this.reportSchemaError("s4s-att-must-appear", new Object[] { content.getLocalName(), SchemaSymbols.ATT_VALUE }, content);
                    }
                    this.fAttrChecker.returnAttrArray(attrs, schemaDoc);
                    content = DOMUtil.getNextSiblingElement(content);
                    continue;
                }
                else {
                    facetsPresent |= currentFacet;
                    if ((boolean)attrs[XSAttributeChecker.ATTIDX_FIXED]) {
                        facetsFixed |= currentFacet;
                    }
                    switch (currentFacet) {
                        case 2: {
                            this.xsFacets.minLength = ((XInt)attrs[XSAttributeChecker.ATTIDX_VALUE]).intValue();
                            hasMinLengthFacet = true;
                            break;
                        }
                        case 4: {
                            this.xsFacets.maxLength = ((XInt)attrs[XSAttributeChecker.ATTIDX_VALUE]).intValue();
                            hasMaxLengthFacet = true;
                            break;
                        }
                        case 64: {
                            this.xsFacets.maxExclusive = (String)attrs[XSAttributeChecker.ATTIDX_VALUE];
                            break;
                        }
                        case 32: {
                            this.xsFacets.maxInclusive = (String)attrs[XSAttributeChecker.ATTIDX_VALUE];
                            break;
                        }
                        case 128: {
                            this.xsFacets.minExclusive = (String)attrs[XSAttributeChecker.ATTIDX_VALUE];
                            break;
                        }
                        case 256: {
                            this.xsFacets.minInclusive = (String)attrs[XSAttributeChecker.ATTIDX_VALUE];
                            break;
                        }
                        case 512: {
                            this.xsFacets.totalDigits = ((XInt)attrs[XSAttributeChecker.ATTIDX_VALUE]).intValue();
                            break;
                        }
                        case 1024: {
                            this.xsFacets.fractionDigits = ((XInt)attrs[XSAttributeChecker.ATTIDX_VALUE]).intValue();
                            break;
                        }
                        case 16: {
                            this.xsFacets.whiteSpace = ((XInt)attrs[XSAttributeChecker.ATTIDX_VALUE]).shortValue();
                            break;
                        }
                        case 1: {
                            this.xsFacets.length = ((XInt)attrs[XSAttributeChecker.ATTIDX_VALUE]).intValue();
                            hasLengthFacet = true;
                            break;
                        }
                    }
                    Element child = DOMUtil.getFirstChildElement(content);
                    XSAnnotationImpl annotation = null;
                    if (child != null && DOMUtil.getLocalName(child).equals(SchemaSymbols.ELT_ANNOTATION)) {
                        annotation = this.traverseAnnotationDecl(child, attrs, false, schemaDoc);
                        child = DOMUtil.getNextSiblingElement(child);
                    }
                    else {
                        String text = DOMUtil.getSyntheticAnnotation(content);
                        if (text != null) {
                            annotation = this.traverseSyntheticAnnotation(content, text, attrs, false, schemaDoc);
                        }
                    }
                    switch (currentFacet) {
                        case 2: {
                            this.xsFacets.minLengthAnnotation = annotation;
                            break;
                        }
                        case 4: {
                            this.xsFacets.maxLengthAnnotation = annotation;
                            break;
                        }
                        case 64: {
                            this.xsFacets.maxExclusiveAnnotation = annotation;
                            break;
                        }
                        case 32: {
                            this.xsFacets.maxInclusiveAnnotation = annotation;
                            break;
                        }
                        case 128: {
                            this.xsFacets.minExclusiveAnnotation = annotation;
                            break;
                        }
                        case 256: {
                            this.xsFacets.minInclusiveAnnotation = annotation;
                            break;
                        }
                        case 512: {
                            this.xsFacets.totalDigitsAnnotation = annotation;
                            break;
                        }
                        case 1024: {
                            this.xsFacets.fractionDigitsAnnotation = annotation;
                            break;
                        }
                        case 16: {
                            this.xsFacets.whiteSpaceAnnotation = annotation;
                            break;
                        }
                        case 1: {
                            this.xsFacets.lengthAnnotation = annotation;
                            break;
                        }
                    }
                    if (child != null) {
                        this.reportSchemaError("s4s-elt-must-match.1", new Object[] { facet, "(annotation?)", DOMUtil.getLocalName(child) }, child);
                    }
                }
            }
            this.fAttrChecker.returnAttrArray(attrs, schemaDoc);
            content = DOMUtil.getNextSiblingElement(content);
        }
        if (enumData != null) {
            facetsPresent |= 0x800;
            this.xsFacets.enumeration = enumData;
            this.xsFacets.enumNSDecls = enumNSDecls;
            this.xsFacets.enumAnnotations = enumAnnotations;
        }
        if (seenPattern) {
            facetsPresent |= 0x8;
            this.xsFacets.pattern = this.fPattern.toString();
            this.xsFacets.patternAnnotations = patternAnnotations;
        }
        this.fPattern.setLength(0);
        if (enumData != null) {
            if (hasLengthFacet) {
                this.checkEnumerationAndLengthInconsistency(baseValidator, enumData, contextNode, getSchemaTypeName(typeDef));
            }
            if (hasMinLengthFacet) {
                this.checkEnumerationAndMinLengthInconsistency(baseValidator, enumData, contextNode, getSchemaTypeName(typeDef));
            }
            if (hasMaxLengthFacet) {
                this.checkEnumerationAndMaxLengthInconsistency(baseValidator, enumData, contextNode, getSchemaTypeName(typeDef));
            }
        }
        return new XSDAbstractTraverser.FacetInfo(this.xsFacets, content, facetsPresent, facetsFixed);
    }
    
    public static String getSchemaTypeName(XSTypeDefinition typeDefn) {
        String typeNameStr = "";
        if ((typeDefn instanceof XSSimpleTypeDefinition)) {
            typeNameStr = ((XSSimpleTypeDecl)typeDefn).getTypeName();
            return typeNameStr;
        }
        typeNameStr = ((XSComplexTypeDecl)typeDefn).getTypeName();
        return typeNameStr;
    }
    
    private void checkEnumerationAndMaxLengthInconsistency(XSSimpleType baseValidator, List<String> enumData, Element contextNode, String typeName) {
        if (SchemaSymbols.URI_SCHEMAFORSCHEMA.equals(baseValidator.getNamespace()) && "hexBinary".equals(baseValidator.getName())) {
            for (int enumIdx = 0; enumIdx < enumData.size(); ++enumIdx) {
                String enumVal = (String)enumData.get(enumIdx);
                if (enumVal.length() / 2 > this.xsFacets.maxLength) {
                    this.reportSchemaWarning("FacetsContradict", new Object[] { enumVal, SchemaSymbols.ELT_MAXLENGTH, typeName }, contextNode);
                }
            }
        }
        else if (SchemaSymbols.URI_SCHEMAFORSCHEMA.equals(baseValidator.getNamespace()) && "base64Binary".equals(baseValidator.getName())) {
            for (int enumIdx = 0; enumIdx < enumData.size(); ++enumIdx) {
                String enumVal = (String)enumData.get(enumIdx);
                byte[] decodedVal = Base64.decode(enumVal);
                if (decodedVal != null && new String(decodedVal).length() > this.xsFacets.maxLength) {
                    this.reportSchemaWarning("FacetsContradict", new Object[] { enumVal, SchemaSymbols.ELT_MAXLENGTH, typeName }, contextNode);
                }
            }
        }
        else {
            for (int enumIdx = 0; enumIdx < enumData.size(); ++enumIdx) {
                String enumVal = (String)enumData.get(enumIdx);
                if (enumVal.length() > this.xsFacets.maxLength) {
                    this.reportSchemaWarning("FacetsContradict", new Object[] { enumVal, SchemaSymbols.ELT_MAXLENGTH, typeName }, contextNode);
                }
            }
        }
    }
    
    private void checkEnumerationAndMinLengthInconsistency(XSSimpleType baseValidator, List<String> enumData, Element contextNode, String typeName) {
        if (SchemaSymbols.URI_SCHEMAFORSCHEMA.equals(baseValidator.getNamespace()) && "hexBinary".equals(baseValidator.getName())) {
            for (int enumIdx = 0; enumIdx < enumData.size(); ++enumIdx) {
                String enumVal = (String)enumData.get(enumIdx);
                if (enumVal.length() / 2 < this.xsFacets.minLength) {
                    this.reportSchemaWarning("FacetsContradict", new Object[] { enumVal, SchemaSymbols.ELT_MINLENGTH, typeName }, contextNode);
                }
            }
        }
        else if (SchemaSymbols.URI_SCHEMAFORSCHEMA.equals(baseValidator.getNamespace()) && "base64Binary".equals(baseValidator.getName())) {
            for (int enumIdx = 0; enumIdx < enumData.size(); ++enumIdx) {
                String enumVal = (String)enumData.get(enumIdx);
                byte[] decodedVal = Base64.decode(enumVal);
                if (decodedVal != null && new String(decodedVal).length() < this.xsFacets.minLength) {
                    this.reportSchemaWarning("FacetsContradict", new Object[] { enumVal, SchemaSymbols.ELT_MINLENGTH, typeName }, contextNode);
                }
            }
        }
        else {
            for (int enumIdx = 0; enumIdx < enumData.size(); ++enumIdx) {
                String enumVal = (String)enumData.get(enumIdx);
                if (enumVal.length() < this.xsFacets.minLength) {
                    this.reportSchemaWarning("FacetsContradict", new Object[] { enumVal, SchemaSymbols.ELT_MINLENGTH, typeName }, contextNode);
                }
            }
        }
    }
    
    private void checkEnumerationAndLengthInconsistency(XSSimpleType baseValidator, List<String> enumData, Element contextNode, String typeName) {
        if (SchemaSymbols.URI_SCHEMAFORSCHEMA.equals(baseValidator.getNamespace()) && "hexBinary".equals(baseValidator.getName())) {
            for (int enumIdx = 0; enumIdx < enumData.size(); ++enumIdx) {
                String enumVal = (String)enumData.get(enumIdx);
                if (enumVal.length() / 2 != this.xsFacets.length) {
                    this.reportSchemaWarning("FacetsContradict", new Object[] { enumVal, SchemaSymbols.ELT_LENGTH, typeName }, contextNode);
                }
            }
        }
        else if (SchemaSymbols.URI_SCHEMAFORSCHEMA.equals(baseValidator.getNamespace()) && "base64Binary".equals(baseValidator.getName())) {
            for (int enumIdx = 0; enumIdx < enumData.size(); ++enumIdx) {
                String enumVal = (String)enumData.get(enumIdx);
                byte[] decodedVal = Base64.decode(enumVal);
                if (decodedVal != null && new String(decodedVal).length() != this.xsFacets.length) {
                    this.reportSchemaWarning("FacetsContradict", new Object[] { enumVal, SchemaSymbols.ELT_LENGTH, typeName }, contextNode);
                }
            }
        }
        else {
            for (int enumIdx = 0; enumIdx < enumData.size(); ++enumIdx) {
                String enumVal = (String)enumData.get(enumIdx);
                if (enumVal.length() != this.xsFacets.length) {
                    this.reportSchemaWarning("FacetsContradict", new Object[] { enumVal, SchemaSymbols.ELT_LENGTH, typeName }, contextNode);
                }
            }
        }
    }
    
    private boolean containsQName(XSSimpleType type) {
        if (type.getVariety() == 1) {
            short primitive = type.getPrimitiveKind();
            return primitive == 18 || primitive == 20;
        }
        if (type.getVariety() == 2) {
            return this.containsQName((XSSimpleType)type.getItemType());
        }
        if (type.getVariety() == 3) {
            XSObjectList members = type.getMemberTypes();
            int i = 0;
            while (i < members.getLength()) {
                if (this.containsQName((XSSimpleType)members.item(i))) {
                    return true;
                }
                ++i;
            }
        }
        return false;
    }
    
    Element traverseAttrsAndAttrGrps(Element firstAttr, XSAttributeGroupDecl attrGrp, XSDocumentInfo schemaDoc, SchemaGrammar grammar, XSComplexTypeDecl enclosingCT) {
        Element child = null;
        XSAttributeGroupDecl tempAttrGrp = null;
        XSAttributeUseImpl tempAttrUse = null;
        XSAttributeUse otherUse = null;
        for (child = firstAttr; child != null; child = DOMUtil.getNextSiblingElement(child)) {
            String childName = DOMUtil.getLocalName(child);
            if (childName.equals(SchemaSymbols.ELT_ATTRIBUTE)) {
                tempAttrUse = this.fSchemaHandler.fAttributeTraverser.traverseLocal(child, schemaDoc, grammar, enclosingCT);
                if (tempAttrUse != null) {
                    if (tempAttrUse.fUse == 2) {
                        attrGrp.addAttributeUse(tempAttrUse);
                    }
                    else {
                        otherUse = attrGrp.getAttributeUseNoProhibited(tempAttrUse.fAttrDecl.getNamespace(), tempAttrUse.fAttrDecl.getName());
                        if (otherUse == null) {
                            String idName2 = attrGrp.addAttributeUse(tempAttrUse);
                            if (idName2 != null) {
                                String code5 = (enclosingCT == null) ? "ag-props-correct.3" : "ct-props-correct.5";
                                String name5 = (enclosingCT == null) ? attrGrp.fName : enclosingCT.getName();
                                this.reportSchemaError(code5, new Object[] { name5, tempAttrUse.fAttrDecl.getName(), idName2 }, child);
                            }
                        }
                        else if (otherUse != tempAttrUse) {
                            String code4 = (enclosingCT == null) ? "ag-props-correct.2" : "ct-props-correct.4";
                            String name4 = (enclosingCT == null) ? attrGrp.fName : enclosingCT.getName();
                            this.reportSchemaError(code4, new Object[] { name4, tempAttrUse.fAttrDecl.getName() }, child);
                        }
                    }
                }
            }
            else if (childName.equals(SchemaSymbols.ELT_ATTRIBUTEGROUP)) {
                tempAttrGrp = this.fSchemaHandler.fAttributeGroupTraverser.traverseLocal(child, schemaDoc, grammar);
                if (tempAttrGrp != null) {
                    XSObjectList attrUseS = tempAttrGrp.getAttributeUses();
                    for (int attrCount = attrUseS.getLength(), i = 0; i < attrCount; ++i) {
                        XSAttributeUseImpl oneAttrUse = (XSAttributeUseImpl)attrUseS.item(i);
                        if (oneAttrUse.fUse == 2) {
                            attrGrp.addAttributeUse(oneAttrUse);
                        }
                        else {
                            otherUse = attrGrp.getAttributeUseNoProhibited(oneAttrUse.fAttrDecl.getNamespace(), oneAttrUse.fAttrDecl.getName());
                            if (otherUse == null) {
                                String idName = attrGrp.addAttributeUse(oneAttrUse);
                                if (idName != null) {
                                    String code2 = (enclosingCT == null) ? "ag-props-correct.3" : "ct-props-correct.5";
                                    String name2 = (enclosingCT == null) ? attrGrp.fName : enclosingCT.getName();
                                    this.reportSchemaError(code2, new Object[] { name2, oneAttrUse.fAttrDecl.getName(), idName }, child);
                                }
                            }
                            else if (oneAttrUse != otherUse) {
                                String code = (enclosingCT == null) ? "ag-props-correct.2" : "ct-props-correct.4";
                                String name = (enclosingCT == null) ? attrGrp.fName : enclosingCT.getName();
                                this.reportSchemaError(code, new Object[] { name, oneAttrUse.fAttrDecl.getName() }, child);
                            }
                        }
                    }
                    if (tempAttrGrp.fAttributeWC != null) {
                        if (attrGrp.fAttributeWC == null) {
                            attrGrp.fAttributeWC = tempAttrGrp.fAttributeWC;
                        }
                        else {
                            attrGrp.fAttributeWC = attrGrp.fAttributeWC.performIntersectionWith(tempAttrGrp.fAttributeWC, attrGrp.fAttributeWC.fProcessContents);
                            if (attrGrp.fAttributeWC == null) {
                                String code3 = (enclosingCT == null) ? "src-attribute_group.2" : "src-ct.4";
                                String name3 = (enclosingCT == null) ? attrGrp.fName : enclosingCT.getName();
                                this.reportSchemaError(code3, new Object[] { name3 }, child);
                            }
                        }
                    }
                }
            }
            else {
                break;
            }
        }
        if (child != null) {
            String childName = DOMUtil.getLocalName(child);
            if (childName.equals(SchemaSymbols.ELT_ANYATTRIBUTE)) {
                XSWildcardDecl tempAttrWC = this.fSchemaHandler.fWildCardTraverser.traverseAnyAttribute(child, schemaDoc, grammar);
                if (attrGrp.fAttributeWC == null) {
                    attrGrp.fAttributeWC = tempAttrWC;
                }
                else {
                    attrGrp.fAttributeWC = tempAttrWC.performIntersectionWith(attrGrp.fAttributeWC, tempAttrWC.fProcessContents);
                    if (attrGrp.fAttributeWC == null) {
                        String code5 = (enclosingCT == null) ? "src-attribute_group.2" : "src-ct.4";
                        String name5 = (enclosingCT == null) ? attrGrp.fName : enclosingCT.getName();
                        this.reportSchemaError(code5, new Object[] { name5 }, child);
                    }
                }
                child = DOMUtil.getNextSiblingElement(child);
            }
        }
        return child;
    }
    
    void reportSchemaError(String key, Object[] args, Element ele) {
        this.fSchemaHandler.reportSchemaError(key, args, ele);
    }
    
    void reportSchemaWarning(String key, Object[] args, Element ele) {
        this.fSchemaHandler.reportSchemaWarning(key, args, ele);
    }
    
    void checkNotationType(String refName, XSTypeDefinition typeDecl, Element elem) {
        if (typeDecl.getTypeCategory() == 16 && ((XSSimpleType)typeDecl).getVariety() == 1 && ((XSSimpleType)typeDecl).getPrimitiveKind() == 20 && (((XSSimpleType)typeDecl).getDefinedFacets() & 0x800) == 0x0) {
            this.reportSchemaError("enumeration-required-notation", new Object[] { typeDecl.getName(), refName, DOMUtil.getLocalName(elem) }, elem);
        }
    }
    
    protected XSParticleDecl checkOccurrences(XSParticleDecl particle, String particleName, Element parent, int allContextFlags, long defaultVals) {
        int min = particle.fMinOccurs;
        int max = particle.fMaxOccurs;
        boolean defaultMin = (defaultVals & (long)(1 << XSAttributeChecker.ATTIDX_MINOCCURS)) != 0x0L;
        boolean defaultMax = (defaultVals & (long)(1 << XSAttributeChecker.ATTIDX_MAXOCCURS)) != 0x0L;
        boolean processingAllEl = (allContextFlags & 0x1) != 0x0;
        boolean processingAllGP = (allContextFlags & 0x8) != 0x0;
        boolean groupRefWithAll = (allContextFlags & 0x2) != 0x0;
        boolean isGroupChild = (allContextFlags & 0x4) != 0x0;
        if (isGroupChild) {
            if (!defaultMin) {
                Object[] args = { particleName, "minOccurs" };
                this.reportSchemaError("s4s-att-not-allowed", args, parent);
                min = 1;
            }
            if (!defaultMax) {
                Object[] args = { particleName, "maxOccurs" };
                this.reportSchemaError("s4s-att-not-allowed", args, parent);
                max = 1;
            }
        }
        if (min == 0 && max == 0) {
            particle.fType = 0;
            return null;
        }
        if (processingAllEl) {
            if (max != 1) {
                this.reportSchemaError("cos-all-limited.2", new Object[] { (max == -1) ? "unbounded" : Integer.toString(max), ((XSElementDecl)particle.fValue).getName() }, parent);
                max = 1;
                if (min > 1) {
                    min = 1;
                }
            }
        }
        else if ((processingAllGP || groupRefWithAll) && max != 1) {
            this.reportSchemaError("cos-all-limited.1.2", null, parent);
            if (min > 1) {
                min = 1;
            }
            max = 1;
        }
        particle.fMinOccurs = min;
        particle.fMaxOccurs = max;
        return particle;
    }
    
    private static String processAttValue(String original) {
        int length = original.length();
        int i = 0;
        while (i < length) {
            char currChar = original.charAt(i);
            if (currChar == '\"' || currChar == '<' || currChar == '&' || currChar == '\t' || currChar == '\n' || currChar == '\r') {
                return escapeAttValue(original, i);
            }
            ++i;
        }
        return original;
    }
    
    private static String escapeAttValue(String original, int from) {
        int length = original.length();
        StringBuilder newVal = new StringBuilder(length);
        newVal.append(original.substring(0, from));
        for (int i = from; i < length; ++i) {
            char currChar = original.charAt(i);
            if (currChar == '\"') {
                newVal.append("&quot;");
            }
            else if (currChar == '<') {
                newVal.append("&lt;");
            }
            else if (currChar == '&') {
                newVal.append("&amp;");
            }
            else if (currChar == '\t') {
                newVal.append("&#x9;");
            }
            else if (currChar == '\n') {
                newVal.append("&#xA;");
            }
            else if (currChar == '\r') {
                newVal.append("&#xD;");
            }
            else {
                newVal.append(currChar);
            }
        }
        return newVal.toString();
    }
    
    static {
        fQNameDV = (XSSimpleType)SchemaGrammar.SG_SchemaNS.getGlobalTypeDecl("QName");
    }
    
    static final class FacetInfo
    {
        final XSFacets facetdata;
        final Element nodeAfterFacets;
        final short fPresentFacets;
        final short fFixedFacets;
        
        FacetInfo(XSFacets facets, Element nodeAfterFacets, short presentFacets, short fixedFacets) {
            this.facetdata = facets;
            this.nodeAfterFacets = nodeAfterFacets;
            this.fPresentFacets = presentFacets;
            this.fFixedFacets = fixedFacets;
        }
    }
}
